from math import cos, sin, radians

import vtk.qt
from qtpyvcp.utilities import logger

LOG = logger.getLogger(__name__)

class ToolActor(vtk.vtkActor):
    def __init__(self, linuxcncDataSource):
        super(ToolActor, self).__init__()
        self._datasource = linuxcncDataSource
        self._tool_table = self._datasource.getToolTable()

        tool = self._tool_table[0]

        if self._datasource.isMachineMetric():
            self.height = 25.4 * 2.0
        else:
            self.height = 2.0

        if self._datasource.isMachineLathe():
            if tool.id == 0 or tool.id == -1:
                polygonSource = vtk.vtkRegularPolygonSource()
                polygonSource.SetNumberOfSides(64)
                polygonSource.SetRadius(0.035)
                polygonSource.SetCenter(0.0, 0.0, 0.0)

                transform = vtk.vtkTransform()
                transform.RotateWXYZ(90, 1, 0, 0)

                transform_filter = vtk.vtkTransformPolyDataFilter()
                transform_filter.SetTransform(transform)
                transform_filter.SetInputConnection(polygonSource.GetOutputPort())
                transform_filter.Update()

                mapper = vtk.vtkPolyDataMapper()
                mapper.SetInputConnection(transform_filter.GetOutputPort())
            else:
                if tool.orientation == 1 and tool.frontangle == 90 and tool.backangle == 90:

                    # Setup four points
                    points = vtk.vtkPoints()
                    points.InsertNextPoint((-tool.xoffset, 0.0, -tool.zoffset))
                    points.InsertNextPoint((-tool.xoffset + 0.5, 0.0, -tool.zoffset))
                    points.InsertNextPoint((-tool.xoffset + 0.5, 0.0, -tool.zoffset - 0.05))
                    points.InsertNextPoint((-tool.xoffset, 0.0, -tool.zoffset - 0.05))

                    # Create the polygon
                    # Create a quad on the four points
                    quad = vtk.vtkQuad()
                    quad.GetPointIds().SetId(0, 0)
                    quad.GetPointIds().SetId(1, 1)
                    quad.GetPointIds().SetId(2, 2)
                    quad.GetPointIds().SetId(3, 3)

                    # Add the polygon to a list of polygons
                    polygons = vtk.vtkCellArray()
                    polygons.InsertNextCell(quad)

                    # Create a PolyData
                    polygonPolyData = vtk.vtkPolyData()
                    polygonPolyData.SetPoints(points)
                    polygonPolyData.SetPolys(polygons)

                    # Create a mapper and actor
                    mapper = vtk.vtkPolyDataMapper()
                    mapper.SetInputData(polygonPolyData)

                elif tool.orientation == 2 and tool.frontangle == 90 and tool.backangle == 90:

                    # Setup four points
                    points = vtk.vtkPoints()
                    points.InsertNextPoint((-tool.xoffset, 0.0, -tool.zoffset))
                    points.InsertNextPoint((-tool.xoffset, 0.0, -tool.zoffset + 0.05))
                    points.InsertNextPoint((-tool.xoffset + 0.5, 0.0, -tool.zoffset + 0.05))
                    points.InsertNextPoint((-tool.xoffset + 0.5, 0.0, -tool.zoffset))

                    # Create the polygon
                    # Create a quad on the four points
                    quad = vtk.vtkQuad()
                    quad.GetPointIds().SetId(0, 0)
                    quad.GetPointIds().SetId(1, 1)
                    quad.GetPointIds().SetId(2, 2)
                    quad.GetPointIds().SetId(3, 3)

                    # Add the polygon to a list of polygons
                    polygons = vtk.vtkCellArray()
                    polygons.InsertNextCell(quad)

                    # Create a PolyData
                    polygonPolyData = vtk.vtkPolyData()
                    polygonPolyData.SetPoints(points)
                    polygonPolyData.SetPolys(polygons)

                    # Create a mapper and actor
                    mapper = vtk.vtkPolyDataMapper()
                    mapper.SetInputData(polygonPolyData)

                elif tool.orientation == 3 and tool.frontangle == 90 and tool.backangle == 90:

                    # Setup four points
                    points = vtk.vtkPoints()
                    points.InsertNextPoint((-tool.xoffset, 0.0, -tool.zoffset))
                    points.InsertNextPoint((-tool.xoffset, 0.0, -tool.zoffset + 0.05))
                    points.InsertNextPoint((-tool.xoffset - 0.5, 0.0, -tool.zoffset + 0.05))
                    points.InsertNextPoint((-tool.xoffset - 0.5, 0.0, -tool.zoffset))

                    # Create the polygon
                    # Create a quad on the four points
                    quad = vtk.vtkQuad()
                    quad.GetPointIds().SetId(0, 0)
                    quad.GetPointIds().SetId(1, 1)
                    quad.GetPointIds().SetId(2, 2)
                    quad.GetPointIds().SetId(3, 3)

                    # Add the polygon to a list of polygons
                    polygons = vtk.vtkCellArray()
                    polygons.InsertNextCell(quad)

                    # Create a PolyData
                    polygonPolyData = vtk.vtkPolyData()
                    polygonPolyData.SetPoints(points)
                    polygonPolyData.SetPolys(polygons)

                    # Create a mapper and actor
                    mapper = vtk.vtkPolyDataMapper()
                    mapper.SetInputData(polygonPolyData)

                elif tool.orientation == 4 and tool.frontangle == 90 and tool.backangle == 90:

                    # Setup four points
                    points = vtk.vtkPoints()
                    points.InsertNextPoint((-tool.xoffset, 0.0, -tool.zoffset))
                    points.InsertNextPoint((-tool.xoffset, 0.0, -tool.zoffset - 0.05))
                    points.InsertNextPoint((-tool.xoffset - 0.5, 0.0, -tool.zoffset - 0.05))
                    points.InsertNextPoint((-tool.xoffset - 0.5, 0.0, -tool.zoffset))

                    # Create the polygon
                    # Create a quad on the four points
                    quad = vtk.vtkQuad()
                    quad.GetPointIds().SetId(0, 0)
                    quad.GetPointIds().SetId(1, 1)
                    quad.GetPointIds().SetId(2, 2)
                    quad.GetPointIds().SetId(3, 3)

                    # Add the polygon to a list of polygons
                    polygons = vtk.vtkCellArray()
                    polygons.InsertNextCell(quad)

                    # Create a PolyData
                    polygonPolyData = vtk.vtkPolyData()
                    polygonPolyData.SetPoints(points)
                    polygonPolyData.SetPolys(polygons)

                    # Create a mapper and actor
                    mapper = vtk.vtkPolyDataMapper()
                    mapper.SetInputData(polygonPolyData)

                elif tool.orientation == 9:

                    radius = tool.diameter / 2

                    # Setup four points
                    points = vtk.vtkPoints()
                    points.InsertNextPoint((-tool.xoffset + radius, 0.0, -tool.zoffset))
                    points.InsertNextPoint((-tool.xoffset + radius, 0.0, -tool.zoffset + 1.0))
                    points.InsertNextPoint((-tool.xoffset - radius, 0.0, -tool.zoffset + 1.0))
                    points.InsertNextPoint((-tool.xoffset - radius, 0.0, -tool.zoffset))

                    # Create the polygon
                    # Create a quad on the four points
                    quad = vtk.vtkQuad()
                    quad.GetPointIds().SetId(0, 0)
                    quad.GetPointIds().SetId(1, 1)
                    quad.GetPointIds().SetId(2, 2)
                    quad.GetPointIds().SetId(3, 3)

                    # Add the polygon to a list of polygons
                    polygons = vtk.vtkCellArray()
                    polygons.InsertNextCell(quad)

                    # Create a PolyData
                    polygonPolyData = vtk.vtkPolyData()
                    polygonPolyData.SetPoints(points)
                    polygonPolyData.SetPolys(polygons)

                    # Create a mapper and actor
                    mapper = vtk.vtkPolyDataMapper()
                    mapper.SetInputData(polygonPolyData)
                else:
                    positive = 1
                    negative = -1

                    if tool.orientation == 1:
                        fa_x_pol = negative
                        fa_z_pol = negative

                        ba_x_pol = negative
                        ba_z_pol = negative

                    elif tool.orientation == 2:
                        fa_x_pol = negative
                        fa_z_pol = positive

                        ba_x_pol = negative
                        ba_z_pol = positive

                    elif tool.orientation == 3:
                        fa_x_pol = positive
                        fa_z_pol = positive

                        ba_x_pol = positive
                        ba_z_pol = positive

                    elif tool.orientation == 4:
                        fa_x_pol = positive
                        fa_z_pol = negative

                        ba_x_pol = positive
                        ba_z_pol = negative

                    elif tool.orientation == 5:
                        fa_x_pol = positive
                        fa_z_pol = negative

                        ba_x_pol = negative
                        ba_z_pol = positive

                    elif tool.orientation == 6:
                        fa_x_pol = negative
                        fa_z_pol = positive

                        ba_x_pol = negative
                        ba_z_pol = negative

                    elif tool.orientation == 7:
                        fa_x_pol = positive
                        fa_z_pol = positive

                        ba_x_pol = negative
                        ba_z_pol = positive

                    elif tool.orientation == 8:
                        fa_x_pol = positive
                        fa_z_pol = positive

                        ba_x_pol = positive
                        ba_z_pol = negative
                    else:
                        fa_x_pol = 0.0
                        fa_z_pol = 0.0

                        ba_x_pol = 0.0
                        ba_z_pol = 0.0

                    A = radians(float(tool.frontangle))
                    B = radians(float(tool.backangle))
                    C = 0.35

                    p1_x = abs(C * sin(A))
                    p1_z = abs(C * cos(A))

                    p2_x = abs(C * sin(B))
                    p2_z = abs(C * cos(B))

                    p1_x_pos = p1_x * fa_x_pol
                    p1_z_pos = p1_z * fa_z_pol

                    p2_x_pos = p2_x * ba_x_pol
                    p2_z_pos = p2_z * ba_z_pol

                    LOG.debug("Drawing Lathe tool id {}".format(tool.id))

                    LOG.debug(
                        "FrontAngle {} Point P1 X = {} P1 Z = {}".format(float(tool.frontangle), p1_x_pos, p1_z_pos))
                    LOG.debug(
                        "BackAngle {} Point P2 X = {} P2 Z = {}".format(float(tool.backangle), p2_x_pos, p2_z_pos))

                    # Setup three points
                    points = vtk.vtkPoints()
                    points.InsertNextPoint((tool.xoffset, 0.0, -tool.zoffset))
                    points.InsertNextPoint((tool.xoffset + p1_x_pos, 0.0, p1_z_pos - tool.zoffset))
                    points.InsertNextPoint((tool.xoffset + p2_x_pos, 0.0, p2_z_pos - tool.zoffset))

                    # Create the polygon
                    polygon = vtk.vtkPolygon()
                    polygon.GetPointIds().SetNumberOfIds(3)  # make a quad
                    polygon.GetPointIds().SetId(0, 0)
                    polygon.GetPointIds().SetId(1, 1)
                    polygon.GetPointIds().SetId(2, 2)

                    # Add the polygon to a list of polygons
                    polygons = vtk.vtkCellArray()
                    polygons.InsertNextCell(polygon)

                    # Create a PolyData
                    polygon_poly_data = vtk.vtkPolyData()
                    polygon_poly_data.SetPoints(points)
                    polygon_poly_data.SetPolys(polygons)

                    transform = vtk.vtkTransform()
                    transform.RotateWXYZ(180, 0, 0, 1)

                    transform_filter = vtk.vtkTransformPolyDataFilter()
                    transform_filter.SetTransform(transform)
                    transform_filter.SetInputData(polygon_poly_data)
                    transform_filter.Update()

                    # Create a mapper
                    mapper = vtk.vtkPolyDataMapper()
                    mapper.SetInputConnection(transform_filter.GetOutputPort())

        else:
            if tool.id == 0 or tool.diameter < .05:
                transform = vtk.vtkTransform()

                source = vtk.vtkConeSource()
                source.SetHeight(self.height / 2)
                source.SetCenter(-self.height / 4 - tool.zoffset, -tool.yoffset, -tool.xoffset)
                source.SetRadius(self.height / 4)
                source.SetResolution(64)
                transform.RotateWXYZ(90, 0, 1, 0)
                transform_filter = vtk.vtkTransformPolyDataFilter()
                transform_filter.SetTransform(transform)
                transform_filter.SetInputConnection(source.GetOutputPort())
                transform_filter.Update()

                # Create a mapper
                mapper = vtk.vtkPolyDataMapper()
                mapper.SetInputConnection(transform_filter.GetOutputPort())
            else:
                transform = vtk.vtkTransform()

                source = vtk.vtkCylinderSource()
                source.SetHeight(self.height / 2)
                source.SetCenter(-tool.xoffset, self.height / 4 - tool.zoffset, tool.yoffset)
                source.SetRadius(tool.diameter / 2)
                source.SetResolution(64)
                transform.RotateWXYZ(90, 1, 0, 0)

                transform_filter = vtk.vtkTransformPolyDataFilter()
                transform_filter.SetTransform(transform)
                transform_filter.SetInputConnection(source.GetOutputPort())
                transform_filter.Update()

                # Create a mapper
                mapper = vtk.vtkPolyDataMapper()
                mapper.SetInputConnection(transform_filter.GetOutputPort())

        self.SetMapper(mapper)

        # Avoid visible backfaces on Linux with some video cards like intel
        # From: https://stackoverflow.com/questions/51357630/vtk-rendering-not-working-as-expected-inside-pyqt?rq=1#comment89720589_51360335
        self.GetProperty().SetBackfaceCulling(1)


import vtk.qt
from qtpyvcp.utilities import logger
from vtk.util.colors import cyan

LOG = logger.getLogger(__name__)

class PathCacheActor(vtk.vtkActor):
    def __init__(self, current_position):
        super(PathCacheActor, self).__init__()
        self.current_position = current_position
        self.index = 0
        self.num_points = 2

        self.points = vtk.vtkPoints()
        self.points.InsertNextPoint(current_position)

        self.lines = vtk.vtkCellArray()
        self.lines.InsertNextCell(1)  # number of points
        self.lines.InsertCellPoint(0)

        self.lines_poligon_data = vtk.vtkPolyData()
        self.polygon_mapper = vtk.vtkPolyDataMapper()
        self.GetProperty().SetColor(cyan)
        self.GetProperty().SetLineWidth(2.5)
        self.GetProperty().SetOpacity(0.5)
        self.SetMapper(self.polygon_mapper)
        
        self.lines_poligon_data.SetPoints(self.points)
        self.lines_poligon_data.SetLines(self.lines)

        self.polygon_mapper.SetInputData(self.lines_poligon_data)
        self.polygon_mapper.Update()

    def add_line_point(self, point):
        self.index += 1

        self.points.InsertNextPoint(point)
        self.points.Modified()

        self.lines.InsertNextCell(self.num_points)
        self.lines.InsertCellPoint(self.index - 1)
        self.lines.InsertCellPoint(self.index)
        self.lines.Modified()


#   Copyright (c) 2018 Kurt Jacobson
#      <kurtcjacobson@gmail.com>
#
#   This file is part of QtPyVCP.
#
#   QtPyVCP is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 2 of the License, or
#   (at your option) any later version.
#
#   QtPyVCP is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with QtPyVCP.  If not, see <http://www.gnu.org/licenses/>.


import linuxcnc
import os

from qtpy import uic
from qtpy.QtWidgets import QDialog

from qtpyvcp.widgets.dialogs.base_dialog import BaseDialog

INIFILE = linuxcnc.ini(os.getenv("INI_FILE_NAME"))
LATHE = bool(INIFILE.find("DISPLAY", "LATHE"))


class ToolEditDialog(BaseDialog):
    def __init__(self, parent=None):
        super(ToolEditDialog, self).__init__(parent=parent, stay_on_top=True)

        self.mill_ui_dialog = 'tool_edit_mill_dialog.ui'
        self.lathe_ui_dialog = 'tool_edit_lathe_dialog.ui'

        if LATHE is True:
            self.dialog_ui = self.lathe_ui_dialog
        else:
            self.dialog_ui = self.mill_ui_dialog

        uic.loadUi(os.path.join(os.path.dirname(__file__), self.dialog_ui), self)



import vtk.qt
from qtpyvcp.utilities import logger

LOG = logger.getLogger(__name__)

class AxesActor(vtk.vtkAxesActor):
    def __init__(self, linuxcncDataSource):
        super(AxesActor, self).__init__()
        self._datasource = linuxcncDataSource
        self._axis_mask = self._datasource.getAxisMask()

        if  self._datasource.isMachineMetric():
            self.length = 20.0
        else:
            self.length = 0.5

        transform = vtk.vtkTransform()
        transform.Translate(0.0, 0.0, 0.0)  # Z up

        self.SetUserTransform(transform)

        self.AxisLabelsOff()
        self.SetShaftTypeToLine()
        self.SetTipTypeToCone()

        # Lathe modes
        if self._axis_mask == 3:
            self.SetTotalLength(self.length, self.length, 0)
        elif self._axis_mask == 5:
            self.SetTotalLength(self.length, 0, self.length)
        elif self._axis_mask == 6:
            self.SetTotalLength(0, self.length, self.length)
        # Mill mode
        else:
            self.SetTotalLength(self.length, self.length, self.length)


import vtk.qt
from axes_actor import AxesActor
from qtpyvcp.utilities import logger

LOG = logger.getLogger(__name__)


class PathActor(vtk.vtkActor):
    def __init__(self, linuxcncDataSource):
        super(PathActor, self).__init__()
        self._datasource = linuxcncDataSource

        self.origin_index = None
        self.origin_cords = None

        if self._datasource.isMachineMetric():
            self.length = 2.5
        else:
            self.length = 0.25

        self.axes_actor = AxesActor(self._datasource)

        if self._datasource.isMachineLathe():
            self.axes_actor.SetTotalLength(self.length, 0, self.length)
        else:
            self.axes_actor.SetTotalLength(self.length, self.length, self.length)

        # Create a vtkUnsignedCharArray container and store the colors in it
        self.colors = vtk.vtkUnsignedCharArray()
        self.colors.SetNumberOfComponents(4)

        self.points = vtk.vtkPoints()
        self.lines = vtk.vtkCellArray()

        self.poly_data = vtk.vtkPolyData()
        self.data_mapper = vtk.vtkPolyDataMapper()

    def set_origin_index(self, index):
        self.origin_index = index

    def get_origin_index(self):
        return self.origin_index

    def set_orgin_coords(self, *cords):
        self.origin_cords = cords

    def get_origin_coords(self):
        return self.origin_cords

    def get_axes_actor(self):
        return self.axes_actor


import sys
from collections import OrderedDict

import vtk
import vtk.qt
from linuxcnc_datasource import LinuxCncDataSource
from path_actor import PathActor
from qtpyvcp.utilities import logger
from qtpyvcp.widgets.display_widgets.vtk_backplot.base_canon import StatCanon

LOG = logger.getLogger(__name__)

COLOR_MAP = {
    'traverse': (200, 35, 35, 255),
    'arcfeed': (110, 110, 255, 255),
    'feed': (210, 210, 255, 255),
    'dwell': (0, 0, 255, 255),
    'user': (0, 100, 255, 255),
}

class VTKCanon(StatCanon):
    def __init__(self, colors=COLOR_MAP, *args, **kwargs):
        super(VTKCanon, self).__init__(*args, **kwargs)
        self._datasource = LinuxCncDataSource()

        self.path_colors = colors
        self.path_actors = OrderedDict()
        self.path_points = OrderedDict()

        self.active_wcs_index = self._datasource.getActiveWcsIndex()

    def comment(self, comment):
        LOG.debug("G-code Comment: {}".format(comment))
        items = comment.lower().split(',', 1)
        if len(items) > 0 and items[0] in ['axis', 'backplot']:
            cmd = items[1].strip()
            if cmd == "hide":
                self.suppress += 1
            elif cmd == "show":
                self.suppress -= 1
            elif cmd == 'stop':
                LOG.info("Backplot generation aborted.")
                raise KeyboardInterrupt

    def message(self, msg):
        LOG.debug("G-code Message: {}".format(msg))

    def set_g5x_offset(self, index, x, y, z, a, b, c, u, v, w):
        new_wcs = index - 1  # this index counts also G53 so we need to do -1
        LOG.debug("---------received wcs change: {}".format(new_wcs))
        if new_wcs not in self.path_actors.keys():
            self.path_actors[new_wcs] = PathActor(self._datasource)
            self.path_points[new_wcs] = list()

        self.active_wcs_index = new_wcs

    def add_path_point(self, line_type, start_point, end_point):
        line = [start_point[:3], end_point[:3]]
        self.path_points.get(self.active_wcs_index).append((line_type, line))

    def draw_lines(self):
        LOG.debug("---------path points size: {}".format(sys.getsizeof(self.path_points)))
        LOG.debug("---------path points length: {}".format(len(self.path_points)))

        # TODO: for some reason, we need to multiply for metric, find out why!
        multiplication_factor = 25.4 if self._datasource.isMachineMetric() else 1

        for wcs_index, data in self.path_points.items():
            index = 0

            path_actor = self.path_actors.get(wcs_index)
            if path_actor is not None:
                for line_type, line_data in data:
                    start_point = line_data[0]
                    end_point = line_data[1]

                    path_actor.points.InsertNextPoint(start_point[0] * multiplication_factor,
                                                      start_point[1] * multiplication_factor,
                                                      start_point[2] * multiplication_factor)

                    path_actor.points.InsertNextPoint(end_point[0] * multiplication_factor,
                                                      end_point[1] * multiplication_factor,
                                                      end_point[2] * multiplication_factor)

                    path_actor.colors.InsertNextTypedTuple(self.path_colors.get(line_type).getRgb()[:4])

                    line = vtk.vtkLine()
                    line.GetPointIds().SetId(0, index)
                    line.GetPointIds().SetId(1, index + 1)

                    path_actor.lines.InsertNextCell(line)

                    index += 2

                # free up memory, lots of it for big files
                self.path_points[wcs_index] = list()

                path_actor.poly_data.SetPoints(path_actor.points)
                path_actor.poly_data.SetLines(path_actor.lines)
                path_actor.poly_data.GetCellData().SetScalars(path_actor.colors)
                path_actor.data_mapper.SetInputData(path_actor.poly_data)
                path_actor.data_mapper.Update()
                path_actor.SetMapper(path_actor.data_mapper)

    def get_path_actors(self):
        return self.path_actors



# import to get true floating point division even if the arguments are ints (ex. 1/2 => 0.5, not 0)
from __future__ import division
from qtpy.QtWidgets import QLineEdit
from qtpyvcp.utilities import logger

from simpleeval import simple_eval

LOG = logger.getLogger(__name__)


class EvalLineEdit(QLineEdit):
    """EvalLineEdit

    This control adds support for evaluating mathematical expressions.

    Any valid expression will be evaluated (ex. -(10+5)*(1.0/2.0) will be evaluated to -2.5).

    Any expression that begins with an operator (except -, see below), will be applied to the current value
    (ex. if current position is 20 and the user enters /2, then the resulting value will be 10).

    Any entry that consists of a single - will invert the sign of the current value.

    NB: If the expression starts with '-' it will be interpreted as negating the value following it.
        This means that you cannot enter -10 and have that value subtract from the current value
        because there is no way to determine if the you wanted to subtract 10, or set the current
        value to -10. In order to subtract from the current value you must use '-=' instead of
        just '-'. This syntax may be used with all operators (ex. +=, /=, *=), but it is only
        required for subtraction.
    """

    def __init__(self, parent=None):
        super(EvalLineEdit, self).__init__(parent)
        self.orig_value = ''
        self.returnPressed.connect(self.evaluate_)

    def focusInEvent(self, event):
        self.orig_value = self.text()

    def evaluate_(self):
        if self.text().strip() == '-':
            # change sign if entry is just a '-'
            self.setText('-' + self.orig_value)
        elif self.text().startswith(('+', '*', '/', '-=')):
            self.setText(self.orig_value + self.text().replace('=', ''))
        try:
            self.setText("{}".format(simple_eval(self.text())))
        except Exception:
            LOG.exception('Error evaluating numeric expression "{}".'.format(self.text()))



import vtk.qt
from qtpyvcp.utilities import logger

LOG = logger.getLogger(__name__)

class MachineActor(vtk.vtkCubeAxesActor):
    def __init__(self, linuxcncDataSource):
        super(MachineActor, self).__init__()
        self._datasource = linuxcncDataSource
        axis = self._datasource.getAxis()
        units = self._datasource.getProgramUnits()

        x_max = axis[0]["max_position_limit"]
        x_min = axis[0]["min_position_limit"]

        y_max = axis[1]["max_position_limit"]
        y_min = axis[1]["min_position_limit"]

        z_max = axis[2]["max_position_limit"]
        z_min = axis[2]["min_position_limit"]

        self.SetBounds(x_min, x_max, y_min, y_max, z_min, z_max)

        self.SetXLabelFormat("%6.3f")
        self.SetYLabelFormat("%6.3f")
        self.SetZLabelFormat("%6.3f")

        self.SetFlyModeToStaticEdges()

        self.GetTitleTextProperty(0).SetColor(1.0, 0.0, 0.0)
        self.GetLabelTextProperty(0).SetColor(1.0, 0.0, 0.0)

        self.GetTitleTextProperty(1).SetColor(0.0, 1.0, 0.0)
        self.GetLabelTextProperty(1).SetColor(0.0, 1.0, 0.0)

        self.GetTitleTextProperty(2).SetColor(0.0, 0.0, 1.0)
        self.GetLabelTextProperty(2).SetColor(0.0, 0.0, 1.0)

        self.SetXUnits(units)
        self.SetYUnits(units)
        self.SetZUnits(units)

        self.DrawXGridlinesOn()
        self.DrawYGridlinesOn()
        self.DrawZGridlinesOn()

        self.SetGridLineLocation(self.VTK_GRID_LINES_FURTHEST)

        self.GetXAxesGridlinesProperty().SetColor(0.0, 0.0, 0.0)
        self.GetYAxesGridlinesProperty().SetColor(0.0, 0.0, 0.0)
        self.GetZAxesGridlinesProperty().SetColor(0.0, 0.0, 0.0)

    def showMachineTicks(self, ticks):
        if ticks:
            self.XAxisTickVisibilityOn()
            self.YAxisTickVisibilityOn()
            self.ZAxisTickVisibilityOn()
        else:
            self.XAxisTickVisibilityOff()
            self.YAxisTickVisibilityOff()
            self.ZAxisTickVisibilityOff()

    def showMachineBounds(self, bounds):
        if bounds:
            self.XAxisVisibilityOn()
            self.YAxisVisibilityOn()
            self.ZAxisVisibilityOn()
        else:
            self.XAxisVisibilityOff()
            self.YAxisVisibilityOff()
            self.ZAxisVisibilityOff()

    def showMachineLabels(self, labels):
        if labels:
            self.XAxisLabelVisibilityOn()
            self.YAxisLabelVisibilityOn()
            self.ZAxisLabelVisibilityOn()
        else:
            self.XAxisLabelVisibilityOff()
            self.YAxisLabelVisibilityOff()
            self.ZAxisLabelVisibilityOff()

    def showGridlines(self, grid):
        if grid:
            self.DrawXGridlinesOn()
            self.DrawYGridlinesOn()
            self.DrawZGridlinesOn()
        else:
            self.DrawXGridlinesOff()
            self.DrawYGridlinesOff()
            self.DrawZGridlinesOff()


"""
QtPyVCP Plugins
---------------

These package level functions provide methods for registering and initializing
plugins, as well as retrieving them for use and terminating them in the proper
order.
"""
import importlib

from collections import OrderedDict

from qtpyvcp.utilities.logger import getLogger
from qtpyvcp.plugins.base_plugins import Plugin, DataPlugin, DataChannel

LOG = getLogger(__name__)

_PLUGINS = OrderedDict()  # Ordered dict so we can initialize/terminate in order


def registerPlugin(plugin_id, plugin_inst):
    """Register a Plugin instance.

    Args:
        plugin_id (str) : The unique name to register the plugin under.
        plugin_inst(plugin_inst) : The plugin instance to register.
    """

    if plugin_id in _PLUGINS:
        LOG.warning("Replacing {} with {} for use with '{}' plugin"
                    .format(_PLUGINS[plugin_id].__class__, plugin_inst.__class__, plugin_id))

    _PLUGINS[plugin_id] = plugin_inst


def registerPluginFromClass(plugin_id, plugin_cls, args=[], kwargs={}):
    """Register a plugin from a class.

    This is primarily used for registering plugins defined in the YAML config.

    .. code-block:: yaml

        data_plugins:
          my_plugin:
            provider: my_package.my_module:MyPluginClass
            args:
              - 10
              - False
            kwargs:
              my_number: 75
              my_string: A string argument

    Args:
        plugin_id (str) : A unique name to register the plugin under.
        plugin_cls (class, str) : A :py:class:`.Plugin` subclass, or a fully
            qualified class spec of format ``package.module:Class`` specifying
            the location of an importable :py:class:`.Plugin` subclass.
        args (list) : Arguments to pass to the plugin's __init__ method.
        kwargs (dict) : Keyword argument to pass to the plugin's __init__ method.

    Returns:
        The plugin instance
    """

    if isinstance(plugin_cls, basestring):
        LOG.debug("Loading plugin '{}' from '{}'".format(plugin_id, plugin_cls))

        modname, sep, clsname = plugin_cls.partition(':')

        try:
            plugin_cls = getattr(importlib.import_module(modname), clsname)
        except Exception:
            LOG.critical("Failed to import data plugin.")
            raise

    assert issubclass(plugin_cls, Plugin), "Not a valid plugin, must be a qtpyvcp.plugins.Plugin subclass."

    try:
        inst = plugin_cls(*args, **kwargs)
        registerPlugin(plugin_id, inst)
        return inst
    except TypeError:
        LOG.critical("Error initializing plugin: {}(*{}, **{})".format(plugin_cls, args, kwargs))
        raise


def getPlugin(plugin_id):
    """Get plugin instance from ID.

    Args:
        plugin_id (str) : The ID of the plugin to retrieve.

    Returns:
        A plugin instance, or None.
    """
    try:
        return _PLUGINS[plugin_id]
    except KeyError:
        LOG.error("Failed to find plugin with ID '%s'", plugin_id)
        return None


def iterPlugins():
    """Returns an iterator for the plugins dict."""
    return _PLUGINS.iteritems()


def initialisePlugins():
    """Initializes all registered plugins.

        Plugins are initialized in the order they were registered in.
        Plugins defined in the YAML file are registered in the order they
        were defined.
    """
    for plugin_id, plugin_inst in _PLUGINS.items():
        LOG.debug("Initializing '%s' plugin", plugin_id)
        plugin_inst.initialise()


def postGuiInitialisePlugins(main_window):
    """Initializes all registered plugins after main window is shown.

        Plugins are initialized in the order they were registered in.
        Plugins defined in the YAML file are registered in the order they
        were defined.
    """
    for plugin_id, plugin_inst in _PLUGINS.items():
        LOG.debug("Post GUI Initializing '%s' plugin", plugin_id)
        plugin_inst.postGuiInitialise(main_window)


def terminatePlugins():
    """Terminates all registered plugins.

        Plugins are terminated in the reverse order they were registered in.
        If an error is encountered while terminating a plugin it will be ignored
        and the remaining plugins will still be terminated.
    """
    # terminate in reverse order, this is to prevent problems
    # when terminating plugins that make use of other plugins.
    for plugin_id, plugin_inst in reversed(_PLUGINS.items()):
        LOG.debug("Terminating '%s' plugin", plugin_id)
        try:
            # try so that other plugins are terminated properly
            # even if one of them fails.
            plugin_inst.terminate()
        except Exception:
            LOG.exception("Error terminating '%s' plugin", plugin_id)





from qtpy.QtWidgets import QApplication, QSpinBox, QDoubleSpinBox, QAbstractButton

from qtpyvcp.app.launcher import _initialize_object_from_dict
from qtpyvcp.utilities.settings import setting
from qtpyvcp.utilities.logger import getLogger
from qtpyvcp.plugins import Plugin


from qtpyvcp import CONFIG

LOG = getLogger(__name__)


class VirtualInputManager(Plugin):
    def __init__(self):
        super(VirtualInputManager, self).__init__()

        self.input_providers = dict()
        self.active_vkb = None

    def activateVirtualInput(self, widget, input_type=None):
        """Activate VKB

        Args:
            widget (QWidget) : The widget for which the VKB should provide input.
            input_type (str) : The type of input that the VKB should provide.
                Input types are specified as either `text` or `number` input,
                with an optional sub-type separated by a colon. For example a
                entry that accepts integer input could request a VKB matching
                an input type of `number:int`. If a VKB does not exist for the
                requested input type the VKB for the base input type will be used.
                If `input_type` is not specified a suitable input type will be
                chosen based on the widget type.
        """

        if input_type is None:
            # determine input type from widget type
            if isinstance(widget, QSpinBox):
                input_type = 'number:int'

            elif isinstance(widget, QDoubleSpinBox):
                input_type = 'number:float'

            elif not isinstance(widget, QAbstractButton) and hasattr(widget, 'text') and hasattr(widget, 'setText'):
                if hasattr(widget, 'isReadOnly') and widget.isReadOnly():
                    return
                try:
                    float(widget.text())
                    input_type = 'number:float'
                except Exception:
                    input_type = 'text:gcode'

            else:
                return

        try:
            vkb = self.input_providers[input_type]
        except KeyError:
            vkb = self.input_providers.get(input_type.split(':')[0])

        if input_type is None:
            return

        vkb.activate(widget)
        self.active_vkb = vkb

    def deactivateVirtualInput(self):
        if self.active_vkb is not None:
            self.active_vkb.hide()
            self.active_vkb = None

    def onFocusChanged(self, old_widget, new_widget):

        # hide any active VKBs
        self.deactivateVirtualInput()

        if new_widget is None or not self.enabled():
            return

        # request VKB for input type
        input_type = new_widget.property('virtualInputType')
        self.activateVirtualInput(new_widget, input_type)

    @setting('virtual-input.enable', False)
    def enabled(obj):
        return obj.value

    def initialise(self):

        virtual_input_providers = CONFIG.get("virtual_input_providers", {})

        for input_type, provider in virtual_input_providers.items():
            obj = _initialize_object_from_dict(provider)
            input_type = input_type.replace('.', ':')
            self.input_providers[input_type] = obj

        app = QApplication.instance()
        app.focusChanged.connect(self.onFocusChanged)

        for vkb in self.input_providers.values():
            vkb.hide()



import os

import vtk.qt
from qtpyvcp.utilities import logger
from qtpyvcp.utilities.settings import getSetting

LOG = logger.getLogger(__name__)
IN_DESIGNER = os.getenv('DESIGNER', False)


class ProgramBoundsActor(vtk.vtkCubeAxesActor):
    def __init__(self, camera, path_actor):
        super(ProgramBoundsActor, self).__init__()

        self.path_actor = path_actor

        """
        for k, v in VTKBackPlot.__dict__.items():
            if "function" in str(v):
                LOG.debug(k)

        for attr_name in dir(VTKBackPlot):
            attr_value = getattr(VTKBackPlot, attr_name)
            LOG.debug(attr_name, attr_value, callable(attr_value))

        LOG.debug(dir(VTKBackPlot))
        testit = getattr(VTKBackPlot, '_enableProgramTicks')
        LOG.debug('enableProgramTicks {}'.format(testit))
        """

        self.SetBounds(self.path_actor.GetBounds())

        self.SetCamera(camera)

        self.SetXLabelFormat("%6.3f")
        self.SetYLabelFormat("%6.3f")
        self.SetZLabelFormat("%6.3f")

        self.SetFlyModeToStaticEdges()

        self.GetTitleTextProperty(0).SetColor(1.0, 0.0, 0.0)
        self.GetLabelTextProperty(0).SetColor(1.0, 0.0, 0.0)

        self.GetTitleTextProperty(1).SetColor(0.0, 1.0, 0.0)
        self.GetLabelTextProperty(1).SetColor(0.0, 1.0, 0.0)

        self.GetTitleTextProperty(2).SetColor(0.0, 0.0, 1.0)
        self.GetLabelTextProperty(2).SetColor(0.0, 0.0, 1.0)

        if not IN_DESIGNER:
            bounds = getSetting('backplot.show-program-bounds')
            self.showProgramBounds(bounds and bounds.value)

            ticks = getSetting('backplot.show-program-ticks')
            self.showProgramTicks(ticks and ticks.value)

            labels = getSetting('backplot.show-program-labels')
            self.showProgramLabels(labels and labels.value)

    def showProgramLabels(self, labels):
        if labels:
            self.XAxisLabelVisibilityOn()
            self.YAxisLabelVisibilityOn()
            self.ZAxisLabelVisibilityOn()
        else:
            self.XAxisLabelVisibilityOff()
            self.YAxisLabelVisibilityOff()
            self.ZAxisLabelVisibilityOff()

    def toggleProgramLabels(self):
        self.showProgramLabels(not self.GetXAxisLabelVisibility())

    def showProgramTicks(self, ticks):
        if ticks:
            self.XAxisTickVisibilityOn()
            self.YAxisTickVisibilityOn()
            self.ZAxisTickVisibilityOn()
        else:
            self.XAxisTickVisibilityOff()
            self.YAxisTickVisibilityOff()
            self.ZAxisTickVisibilityOff()

    def toggleProgramTicks(self):
        self.showProgramTicks(not self.GetXAxisTickVisibility())

    def showProgramBounds(self, bounds):
        if bounds:
            self.XAxisVisibilityOn()
            self.YAxisVisibilityOn()
            self.ZAxisVisibilityOn()
        else:
            self.XAxisVisibilityOff()
            self.YAxisVisibilityOff()
            self.ZAxisVisibilityOff()

    def toggleProgramBounds(self):
        self.showProgramBounds(not self.GetXAxisVisibility())


import linuxcnc
import os
from collections import OrderedDict
from operator import add
import time

import vtk
import vtk.qt
from qtpy.QtCore import Property, Slot
from qtpy.QtGui import QColor

# Fix poligons not drawing correctly on some GPU
# https://stackoverflow.com/questions/51357630/vtk-rendering-not-working-as-expected-inside-pyqt?rq=1

vtk.qt.QVTKRWIBase = "QGLWidget"

# Fix end

from vtk.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor

from qtpyvcp.widgets import VCPWidget
from qtpyvcp.utilities import logger
from qtpyvcp.utilities.settings import connectSetting

from base_backplot import BaseBackPlot
from axes_actor import AxesActor
from machine_actor import MachineActor
from tool_actor import ToolActor
from path_cache_actor import PathCacheActor
from program_bounds_actor import ProgramBoundsActor
from vtk_canon import VTKCanon
from linuxcnc_datasource import LinuxCncDataSource

LOG = logger.getLogger(__name__)

IN_DESIGNER = os.getenv('DESIGNER', False)
NUMBER_OF_WCS = 9

# turn on antialiasing
from PyQt5.QtOpenGL import QGLFormat
f = QGLFormat()
f.setSampleBuffers(True)
QGLFormat.setDefaultFormat(f)

class VTKBackPlot(QVTKRenderWindowInteractor, VCPWidget, BaseBackPlot):
    def __init__(self, parent=None):
        super(VTKBackPlot, self).__init__(parent)
        LOG.debug("---------using refactored vtk code")
        self._datasource = LinuxCncDataSource()

        self.parent = parent
        self.ploter_enabled = True
        self.touch_enabled = False
        self.programViewWhenLoadingProgram = False
        self.pan_mode = False
        self.line = None
        self._last_filename = str()
        self.rotating = 0
        self.panning = 0
        self.zooming = 0

        # assume that we are standing upright and compute azimuth around that axis
        self.natural_view_up = (0, 0, 1)

        # properties
        self._background_color = QColor(0, 0, 0)
        self._background_color2 = QColor(0, 0, 0)
        self._enableProgramTicks = True
        
        self._default_traverse_color = QColor(200, 35, 35, 255)
        self._default_arcfeed_color = QColor(110, 110, 255, 255)
        self._default_feed_color = QColor(210, 210, 255, 255)
        self._default_dwell_color = QColor(0, 0, 255, 255)
        self._default_user_color = QColor(0, 100, 255, 255)
        
        self._traverse_color = self._default_traverse_color
        self._arcfeed_color = self._default_arcfeed_color
        self._feed_color = self._default_feed_color
        self._dwel_color = self._default_dwell_color
        self._user_color = self._default_user_color

        

        self.active_wcs_index = self._datasource.getActiveWcsIndex()
        self.wcs_offsets = self._datasource.getWcsOffsets()
        self.active_wcs_offset = self._datasource.getActiveWcsOffsets()
        self.g92_offset = self._datasource.getG92_offset()

        LOG.debug("---------active_wcs_index {}".format(self.active_wcs_index))
        LOG.debug("---------active_wcs_offset {}".format(self.active_wcs_offset))
        LOG.debug("---------wcs_offsets {}".format(self.wcs_offsets))

        self.original_g5x_offset = [0.0] * NUMBER_OF_WCS
        self.original_g92_offset = [0.0] * NUMBER_OF_WCS

        self.spindle_position = (0.0, 0.0, 0.0)
        self.spindle_rotation = (0.0, 0.0, 0.0)
        self.tooltip_position = (0.0, 0.0, 0.0)

        self.camera = vtk.vtkCamera()
        self.camera.ParallelProjectionOn()

        if self._datasource.isMachineMetric():
            self.position_mult = 1000 #500 here works for me
            self.clipping_range_near = 0.01
            self.clipping_range_far = 10000.0 #TODO: check this value
        else:
            self.position_mult = 100
            self.clipping_range_near = 0.001
            self.clipping_range_far = 1000.0 #TODO: check this value

        self.camera.SetClippingRange(self.clipping_range_near, self.clipping_range_far)
        self.renderer = vtk.vtkRenderer()
        self.renderer.SetActiveCamera(self.camera)

        self.renderer_window = self.GetRenderWindow()
        self.renderer_window.AddRenderer(self.renderer)

        # self.nav_style = vtk.vtkInteractorStyleTrackballCamera()
        self.nav_style = vtk.vtkInteractorStyleMultiTouchCamera() if self.touch_enabled else None
           
        self.interactor = self.renderer_window.GetInteractor()
        self.interactor.SetInteractorStyle(self.nav_style)
        self.interactor.SetRenderWindow(self.renderer_window)
        
        self.machine_actor = MachineActor(self._datasource)
        self.machine_actor.SetCamera(self.camera)

        self.axes_actor = AxesActor(self._datasource)

        LOG.debug("---------translate1: {}".format(self.active_wcs_offset[:3]))
        LOG.debug("---------active_wcs_offset: {}".format(self.active_wcs_offset))

        transform = vtk.vtkTransform()
        transform.Translate(*self.active_wcs_offset[:3])
        transform.RotateZ(self.active_wcs_offset[9])
        self.axes_actor.SetUserTransform(transform)

        self.path_cache_actor = PathCacheActor(self.tooltip_position)
        self.tool_actor = ToolActor(self._datasource)

        self.offset_axes = OrderedDict()
        self.program_bounds_actors = OrderedDict()
        self.show_program_bounds = bool()
    
        
        # Add the observers to watch for particular events. These invoke Python functions.
        self.interactor.AddObserver("LeftButtonPressEvent", self.button_event)
        self.interactor.AddObserver("LeftButtonReleaseEvent", self.button_event)
        self.interactor.AddObserver("MiddleButtonPressEvent", self.button_event)
        self.interactor.AddObserver("MiddleButtonReleaseEvent", self.button_event)
        self.interactor.AddObserver("RightButtonPressEvent", self.button_event)
        self.interactor.AddObserver("RightButtonReleaseEvent", self.button_event)
        self.interactor.AddObserver("MouseMoveEvent", self.mouse_move)
        self.interactor.AddObserver("KeyPressEvent", self.keypress)
        self.interactor.AddObserver("MouseWheelForwardEvent", self.mouse_scroll_forward)
        self.interactor.AddObserver("MouseWheelBackwardEvent", self.mouse_scroll_backward)

        self.interactor.Initialize()
        self.renderer_window.Render()
        self.interactor.Start()

        # Add the observers to watch for particular events. These invoke Python functions.
        self._datasource.programLoaded.connect(self.load_program)
        self._datasource.positionChanged.connect(self.update_position)
        self._datasource.motionTypeChanged.connect(self.motion_type)
        self._datasource.g5xOffsetChanged.connect(self.update_g5x_offset)
        self._datasource.g92OffsetChanged.connect(self.update_g92_offset)
        self._datasource.offsetTableChanged.connect(self.on_offset_table_changed)
        self._datasource.activeOffsetChanged.connect(self.update_active_wcs)
        self._datasource.toolTableChanged.connect(self.update_tool)
        self._datasource.toolOffsetChanged.connect(self.update_tool)
        # self.status.g5x_index.notify(self.update_g5x_index)

        # view settings
        connectSetting('backplot.show-grid', self.showGrid)
        connectSetting('backplot.show-program-bounds', self.showProgramBounds)
        connectSetting('backplot.show-program-labels', self.showProgramLabels)
        connectSetting('backplot.show-program-ticks', self.showProgramTicks)
        connectSetting('backplot.show-machine-bounds', self.showMachineBounds)
        connectSetting('backplot.show-machine-labels', self.showMachineLabels)
        connectSetting('backplot.show-machine-ticks', self.showMachineTicks)
        connectSetting('backplot.perspective-view', self.viewPerspective)
        connectSetting('backplot.view', self.setView)
        connectSetting('backplot.multitool-colors', self.showMultiColorPath)
        

    def initialize(self):
        self.path_colors = {'traverse': self._traverse_color,
                       'arcfeed': self._arcfeed_color,
                       'feed': self._feed_color,
                       'dwell': QColor(0, 0, 255, 255),
                       'user': QColor(0, 100, 255, 255)
                       }
        
        if not IN_DESIGNER:

            self.canon = VTKCanon(colors=self.path_colors)
            self.path_actors = self.canon.get_path_actors()

            for wcs_index, path_actor in self.path_actors.items():
                current_offsets = self.wcs_offsets[wcs_index]

                LOG.debug("---------wcs_offsets: {}".format(self.wcs_offsets))
                LOG.debug("---------wcs_index: {}".format(wcs_index))
                LOG.debug("---------current_offsets: {}".format(current_offsets))

                actor_transform = vtk.vtkTransform()
                actor_transform.Translate(*current_offsets[:3])
                actor_transform.RotateZ(current_offsets[9])

                path_actor.SetUserTransform(actor_transform)
                path_actor.SetPosition(*current_offsets[:3])

                LOG.debug("---------current_position: {}".format(*current_offsets[:3]))

                program_bounds_actor = ProgramBoundsActor(self.camera, path_actor)

                axes = path_actor.get_axes_actor()

                self.offset_axes[wcs_index] = axes
                self.program_bounds_actors[wcs_index] = program_bounds_actor

                self.renderer.AddActor(axes)
                self.renderer.AddActor(program_bounds_actor)
                self.renderer.AddActor(path_actor)

        self.renderer.AddActor(self.tool_actor)
        self.renderer.AddActor(self.machine_actor)
        self.renderer.AddActor(self.axes_actor)
        self.renderer.AddActor(self.path_cache_actor)

        self.renderer.ResetCamera()


    # Handle the mouse button events.
    def button_event(self, obj, event):
        LOG.debug("button event {}".format(event))

        if event == "LeftButtonPressEvent":
            if self.pan_mode is True:
                self.panning = 1
            else:
                self.rotating = 1

        elif event == "LeftButtonReleaseEvent":
            if self.pan_mode is True:
                self.panning = 0
            else:
                self.rotating = 0

        elif event == "MiddleButtonPressEvent":
            if self.pan_mode is True:
                self.rotating = 1
            else:
                self.panning = 1

        elif event == "MiddleButtonReleaseEvent":
            if self.pan_mode is True:
                self.rotating = 0
            else:
                self.panning = 0

        elif event == "RightButtonPressEvent":
            self.zooming = 1
        elif event == "RightButtonReleaseEvent":
            self.zooming = 0

    def mouse_scroll_backward(self, obj, event):
        self.zoomOut()

    def mouse_scroll_forward(self, obj, event):
        self.zoomIn()

    # General high-level logic
    def mouse_move(self, obj, event):
        lastXYpos = self.interactor.GetLastEventPosition()
        lastX = lastXYpos[0]
        lastY = lastXYpos[1]

        xypos = self.interactor.GetEventPosition()
        x = xypos[0]
        y = xypos[1]

        center = self.renderer_window.GetSize()
        centerX = center[0] / 2.0
        centerY = center[1] / 2.0

        if self.rotating:
            if self._datasource.isMachineLathe():
                self.pan(self.renderer, self.camera, x, y, lastX, lastY, centerX, centerY)
            else:
                self.rotate(self.renderer, self.camera, x, y, lastX, lastY, centerX, centerY)
        elif self.panning:
            self.pan(self.renderer, self.camera, x, y, lastX, lastY, centerX, centerY)
        elif self.zooming:
            self.dolly(self.renderer, self.camera, x, y, lastX, lastY, centerX, centerY)

    def keypress(self, obj, event):
        key = obj.GetKeySym()
        if key == 'w' or key == 's':
            self._setRepresentation(key)

    # Functions that translate the events into camera motions.

    # This one is associated with the left mouse button. It translates x
    # and y relative motions into camera azimuth and elevation commands.
    def rotate(self, renderer, camera, x, y, lastX, lastY, centerX, centerY):
        self.natural_azimuth(camera, lastX - x)
        camera.Elevation(lastY - y)
        camera.OrthogonalizeViewUp()
        camera.SetClippingRange(self.clipping_range_near, self.clipping_range_far)
        self.renderer_window.Render()
        # self.renderer.ResetCamera()
        self.interactor.ReInitialize()

    # Change azimuth around natural view up vector
    def natural_azimuth(self, camera, angle):
        fp = self.camera.GetFocalPoint()

        t = vtk.vtkTransform()
        t.Translate(fp[0], fp[1], fp[2])
        t.RotateWXYZ(angle, self.natural_view_up)
        t.Translate(-fp[0], -fp[1], -fp[2])
        camera.ApplyTransform(t)

    # Pan translates x-y motion into translation of the focal point and position.
    def pan(self, renderer, camera, x, y, lastX, lastY, centerX, centerY):
        FPoint = camera.GetFocalPoint()
        FPoint0 = FPoint[0]
        FPoint1 = FPoint[1]
        FPoint2 = FPoint[2]

        PPoint = camera.GetPosition()
        PPoint0 = PPoint[0]
        PPoint1 = PPoint[1]
        PPoint2 = PPoint[2]

        renderer.SetWorldPoint(FPoint0, FPoint1, FPoint2, 1.0)
        renderer.WorldToDisplay()
        DPoint = renderer.GetDisplayPoint()
        focalDepth = DPoint[2]

        APoint0 = centerX + (x - lastX)
        APoint1 = centerY + (y - lastY)

        renderer.SetDisplayPoint(APoint0, APoint1, focalDepth)
        renderer.DisplayToWorld()
        RPoint = renderer.GetWorldPoint()
        RPoint0 = RPoint[0]
        RPoint1 = RPoint[1]
        RPoint2 = RPoint[2]
        RPoint3 = RPoint[3]

        if RPoint3 != 0.0:
            RPoint0 = RPoint0 / RPoint3
            RPoint1 = RPoint1 / RPoint3
            RPoint2 = RPoint2 / RPoint3

        camera.SetFocalPoint((FPoint0 - RPoint0) / 1.0 + FPoint0,
                             (FPoint1 - RPoint1) / 1.0 + FPoint1,
                             (FPoint2 - RPoint2) / 1.0 + FPoint2)

        camera.SetPosition((FPoint0 - RPoint0) / 1.0 + PPoint0,
                           (FPoint1 - RPoint1) / 1.0 + PPoint1,
                           (FPoint2 - RPoint2) / 1.0 + PPoint2)

        self.renderer_window.Render()

    # Dolly converts y-motion into a camera dolly commands.
    def dolly(self, renderer, camera, x, y, lastX, lastY, centerX, centerY):
        dollyFactor = pow(1.02, (0.5 * (y - lastY)))
        if camera.GetParallelProjection():
            parallelScale = camera.GetParallelScale() * dollyFactor
            camera.SetParallelScale(parallelScale)
        else:
            camera.Dolly(dollyFactor)
            renderer.ResetCameraClippingRange()

        self.renderer_window.Render()

    # Surface sets the representation of all actors to surface or wireframe.
    def _setRepresentation(self, keyPressed):
        actors = self.renderer.GetActors()
        actors.InitTraversal()
        actor = actors.GetNextItem()
        while actor:
            if keyPressed == 's':
                # sets the representation of all actors to surface.
                actor.GetProperty().SetRepresentationToSurface()
            elif keyPressed == 'w':
                # sets the representation of all actors to wireframe.
                actor.GetProperty().SetRepresentationToWireframe()
            actor = actors.GetNextItem()
        self.renderer_window.Render()

    def tlo(self, tlo):
        LOG.debug(tlo)

    @Slot()
    def reload_program(self, *args, **kwargs):
        LOG.debug("reload_program")
        self.load_program(self._last_filename)

    def load_program(self, fname=None):
        LOG.debug("-------load_program")

        # Cleanup the scene, remove any previous actors if any
        for wcs_index, actor in self.path_actors.items():
            LOG.debug("-------load_program wcs_index: {}".format(wcs_index))
            axes_actor = actor.get_axes_actor()
            program_bounds_actor = self.program_bounds_actors[wcs_index]

            self.renderer.RemoveActor(axes_actor)
            self.renderer.RemoveActor(actor)
            self.renderer.RemoveActor(program_bounds_actor)

        self.path_actors.clear()
        self.offset_axes.clear()
        self.program_bounds_actors.clear()

        start_time = time.time()

        if fname:
            # create the object which handles the canonical motion callbacks
            # (straight_feed, straight_traverse, arc_feed, rigid_tap, etc.)
            self.canon = VTKCanon(colors=self.path_colors)
            self.load(fname)
        else:
            return

        LOG.debug("-------Load time %s seconds ---" % (time.time() - start_time))

        self.canon.draw_lines()

        LOG.debug("-------Draw time %s seconds ---" % (time.time() - start_time))
        self.path_actors = self.canon.get_path_actors()

        for wcs_index, actor in self.path_actors.items():
            LOG.debug("---------wcs_offsets: {}".format(self.wcs_offsets))
            LOG.debug("---------wcs_index: {}".format(wcs_index))

            current_offsets = self.wcs_offsets[wcs_index]
            LOG.debug("---------current_offsets: {}".format(current_offsets))

            actor_transform = vtk.vtkTransform()
            actor_transform.Translate(*current_offsets[:3])
            actor_transform.RotateZ(current_offsets[9])

            actor.SetUserTransform(actor_transform)
            #actor.SetPosition(path_position[:3])

            LOG.debug("---------current_position: {}".format(*current_offsets[:3]))

            program_bounds_actor = ProgramBoundsActor(self.camera, actor)
            program_bounds_actor.showProgramBounds(self.show_program_bounds)

            axes = actor.get_axes_actor()

            self.offset_axes[wcs_index] = axes
            self.program_bounds_actors[wcs_index] = program_bounds_actor

            axes.SetUserTransform(actor_transform) #TODO: not sure if this is needed

            self.renderer.AddActor(axes)
            self.renderer.AddActor(program_bounds_actor)
            self.renderer.AddActor(actor)

        self.renderer.AddActor(self.axes_actor)
        self.renderer_window.Render()
        if self.setProgramViewWhenLoadingProgram:
            self.setViewProgram()

    def motion_type(self, value):
        LOG.debug("-----motion_type is: {}".format(value))
        if value == linuxcnc.MOTION_TYPE_TOOLCHANGE:
            self.update_tool()

    def update_position(self, position):  # the tool movement

        self.spindle_position = position[:3]
        self.spindle_rotation = position[3:6]

        tool_transform = vtk.vtkTransform()
        tool_transform.Translate(*self.spindle_position)
        tool_transform.RotateX(-self.spindle_rotation[0])
        tool_transform.RotateY(-self.spindle_rotation[1])
        tool_transform.RotateZ(-self.spindle_rotation[2])

        self.tool_actor.SetUserTransform(tool_transform)

        tlo = self._datasource.getToolOffset()
        self.tooltip_position = [pos - tlo for pos, tlo in zip(self.spindle_position, tlo[:3])]

        # self.spindle_actor.SetPosition(self.spindle_position)
        # self.tool_actor.SetPosition(self.spindle_position)
        self.path_cache_actor.add_line_point(self.tooltip_position)
        self.renderer_window.Render()

    def on_offset_table_changed(self, table):
        LOG.debug("on_offset_table_changed")
        self.wcs_offsets = table

    def update_g5x_offset(self, offset):
        LOG.debug("--------update_g5x_offset {}".format(offset))

        transform = vtk.vtkTransform()
        transform.Translate(*offset[:3])
        transform.RotateZ(offset[9])

        self.axes_actor.SetUserTransform(transform)

        for wcs_index, path_actor in self.path_actors.items():

            old_program_bounds_actor = self.program_bounds_actors[wcs_index]
            self.renderer.RemoveActor(old_program_bounds_actor)

            axes = path_actor.get_axes_actor()

            LOG.debug("--------wcs_index: {}, active_wcs_index: {}".format(wcs_index, self.active_wcs_index))

            if wcs_index == self.active_wcs_index:
                path_transform = vtk.vtkTransform()
                path_transform.Translate(*offset[:3])
                path_transform.RotateZ(offset[9])

                axes.SetUserTransform(path_transform)
                path_actor.SetUserTransform(path_transform)

            program_bounds_actor = ProgramBoundsActor(self.camera, path_actor)
            program_bounds_actor.showProgramBounds(self.show_program_bounds)

            self.renderer.AddActor(program_bounds_actor)

            self.program_bounds_actors[wcs_index] = program_bounds_actor

        self.interactor.ReInitialize()
        self.renderer_window.Render()

    def update_active_wcs(self, wcs_index):
        self.active_wcs_index = wcs_index
        LOG.debug("--------update_active_wcs index: {}".format(wcs_index))
        LOG.debug("--------self.wcs_offsets: {}".format(self.wcs_offsets))

        position = self.wcs_offsets[wcs_index]
        LOG.debug("--------position: {}".format(position))

        transform = vtk.vtkTransform()
        transform.Translate(*position[:3])
        transform.RotateZ(position[9])

        self.axes_actor.SetUserTransform(transform)

        self.interactor.ReInitialize()
        self.renderer_window.Render()

    def update_g92_offset(self, g92_offset):
        LOG.debug("---------update_g92_offset: {}".format(g92_offset))
        if self._datasource.isModeMdi() or self._datasource.isModeAuto():
            self.g92_offset = g92_offset

            path_offset = list(map(add, self.g92_offset, self.original_g92_offset))
            LOG.debug("---------path_offset: {}".format(path_offset))

            for wcs_index, actor in self.path_actors.items():

                old_program_bounds_actor = self.program_bounds_actors[wcs_index]
                self.renderer.RemoveActor(old_program_bounds_actor)
                # determine change in g92 offset since path was drawn

                new_path_position = list(map(add, self.wcs_offsets[wcs_index][:9], path_offset))
                LOG.debug("---------new_path_position: {}".format(path_offset))

                axes = actor.get_axes_actor()

                path_transform = vtk.vtkTransform()
                path_transform.Translate(*new_path_position[:3])

                self.axes_actor.SetUserTransform(path_transform)
                axes.SetUserTransform(path_transform)
                actor.SetUserTransform(path_transform)

                program_bounds_actor = ProgramBoundsActor(self.camera, actor)
                program_bounds_actor.showProgramBounds(self.show_program_bounds)

                self.renderer.AddActor(program_bounds_actor)

                self.program_bounds_actors[wcs_index] = program_bounds_actor

            self.interactor.ReInitialize()
            self.renderer_window.Render()

    def update_tool(self):

        LOG.debug("update_tool")

        self.renderer.RemoveActor(self.tool_actor)

        self.tool_actor = ToolActor(self._datasource)

        tool_transform = vtk.vtkTransform()
        tool_transform.Translate(*self.spindle_position)
        tool_transform.RotateX(-self.spindle_rotation[0])
        tool_transform.RotateY(-self.spindle_rotation[1])
        tool_transform.RotateZ(-self.spindle_rotation[2])

        self.tool_actor.SetUserTransform(tool_transform)

        self.renderer.AddActor(self.tool_actor)

        self.renderer_window.Render()

    @Slot(bool)
    @Slot(object)
    def viewPerspective(self, persp):
        if persp:
            self.setViewPersp()
        else:
            self.setViewOrtho()

    @Slot()
    def setViewOrtho(self):
        self.camera.ParallelProjectionOn()
        # self.renderer.ResetCamera()
        self.interactor.ReInitialize()

    @Slot()
    def setViewPersp(self):
        self.camera.ParallelProjectionOff()
        # self.renderer.ResetCamera()
        self.interactor.ReInitialize()

    @Slot(int)
    @Slot(str)
    @Slot(object)
    def setView(self, view):
        if isinstance(view, int):
            view = ['X', 'XZ', 'XZ2', 'Y', 'Z', 'Z2', 'P'][view]

        view = view.upper()
        LOG.debug("Setting view to: %s", view)

        if view == 'X':
            self.setViewX()
        elif view == 'XZ':
            self.setViewXZ()
        elif view == 'XZ2':
            self.setViewXZ2()
        elif view == 'Y':
            self.setViewY()
        elif view == 'Z':
            self.setViewZ()
        elif view == 'Z2':
            self.setViewZ2()
        elif view == 'P':
            self.setViewP()

    @Slot()
    def setViewP(self):
        self.active_view = 'P'
        position = self.wcs_offsets[self.active_wcs_index]
        self.camera.SetPosition(self.position_mult, -self.position_mult, self.position_mult)
        self.camera.SetFocalPoint(position[:3])
        self.camera.SetViewUp(0, 0, 1)
        self.__doCommonSetViewWork()

    @Slot()
    def setViewX(self):
        self.active_view = 'X'
        position = self.wcs_offsets[self.active_wcs_index]
        self.camera.SetPosition(position[0], position[1] - self.position_mult, position[2])
        self.camera.SetFocalPoint(position[:3])
        self.camera.SetViewUp(0, 0, 1)
        self.__doCommonSetViewWork()

    @Slot()
    def setViewXZ(self):
        self.active_view = 'XZ'
        position = self.wcs_offsets[self.active_wcs_index]
        self.camera.SetPosition(position[0], position[1] + self.position_mult, position[2])
        self.camera.SetFocalPoint(position[:3])
        self.camera.SetViewUp(1, 0, 0)
        self.__doCommonSetViewWork()

    @Slot()
    def setViewXZ2(self):
        self.active_view = 'XZ2'
        position = self.wcs_offsets[self.active_wcs_index]
        self.camera.SetPosition(position[0], position[1] - self.position_mult, position[2])
        self.camera.SetFocalPoint(position[:3])
        self.camera.SetViewUp(-1, 0, 0)
        self.__doCommonSetViewWork()

    @Slot()
    def setViewY(self):
        self.active_view = 'Y'
        position = self.wcs_offsets[self.active_wcs_index]
        self.camera.SetPosition(position[0] + self.position_mult, position[1], position[2])
        self.camera.SetFocalPoint(position[:3])
        self.camera.SetViewUp(0, 0, 1)
        self.__doCommonSetViewWork()

    @Slot()
    def setViewZ(self):
        self.active_view = 'Z'
        position = self.wcs_offsets[self.active_wcs_index]
        self.camera.SetPosition(position[0], position[1], position[2] + self.position_mult)
        self.camera.SetFocalPoint(position[:3])
        self.camera.SetViewUp(0, 1, 0)
        self.__doCommonSetViewWork()

    @Slot()
    def setViewZ2(self):
        position = self.wcs_offsets[self.active_wcs_index]
        self.camera.SetPosition(position[0], position[1], position[2] + self.position_mult)
        self.camera.SetFocalPoint(position[:3])
        self.camera.SetViewUp(1, 0, 0)
        self.__doCommonSetViewWork()

    @Slot()
    def setViewMachine(self):
        LOG.debug('-----setViewMachine')
        machine_bounds = self.machine_actor.GetBounds()
        LOG.debug('-----machine_bounds: {}'.format(machine_bounds))

        machine_center = ((machine_bounds[0] + machine_bounds[1]) / 2,
                          (machine_bounds[2] + machine_bounds[3]) / 2,
                          (machine_bounds[4] + machine_bounds[5]) / 2
                          )

        LOG.debug('-----machine_center: {}'.format(machine_center))

        self.camera = self.renderer.GetActiveCamera()

        self.camera.SetFocalPoint(machine_center[0],
                                  machine_center[1],
                                  machine_center[2])

        self.camera.SetPosition(machine_center[0] + self.position_mult,
                                -(machine_center[1] + self.position_mult),
                                machine_center[2] + self.position_mult)

        self.camera.SetViewUp(0, 0, 1)
        self.__doCommonSetViewWork()

    @Slot()
    def setViewProgram(self):
        LOG.debug('-----setViewProgram')
        program_bounds = self.program_bounds_actors[self.active_wcs_index].GetBounds()
        LOG.debug('-----program_bounds: {}'.format(program_bounds))

        program_center = ((program_bounds[0] + program_bounds[1]) / 2,
                          (program_bounds[2] + program_bounds[3]) / 2,
                          (program_bounds[4] + program_bounds[5]) / 2)

        LOG.debug('-----program_center: {}'.format(program_center))

        self.camera = self.renderer.GetActiveCamera()
        self.camera.SetFocalPoint(program_center[0],
                                  program_center[1],
                                  program_center[2])

        self.camera.SetPosition(program_center[0] + self.position_mult,
                                -(program_center[1] + self.position_mult),
                                program_center[2] + self.position_mult)

        x_dist = abs(program_bounds[0] - program_bounds[1])
        y_dist = abs(program_bounds[2] - program_bounds[3])
        z_dist = abs(program_bounds[4] - program_bounds[5])

        LOG.debug('-----x_dist: {}'.format(x_dist))
        LOG.debug('-----y_dist: {}'.format(y_dist))
        LOG.debug('-----z_dist: {}'.format(z_dist))

        scale = max(x_dist, y_dist, z_dist)

        self.camera.SetParallelScale(scale)
        self.camera.SetViewUp(0, 0, 1)
        self.__doCommonSetViewWork()
        self.clearLivePlot()

    @Slot()
    def setViewPath(self):
        LOG.debug('-----setViewPath')
        position = self.wcs_offsets[self.active_wcs_index]
        self.camera.SetPosition(position[0] + self.position_mult,
                                -(position[1] + self.position_mult),
                                position[2] + self.position_mult)
        self.camera.SetFocalPoint(position[:3])
        self.camera.SetViewUp(0, 0, 1)
        self.__doCommonSetViewWork()

    def __doCommonSetViewWork(self):
        # This is common logic for all setView**** methods.
        self.camera.SetClippingRange(self.clipping_range_near, self.clipping_range_far)
        self.renderer_window.Render()
        self.interactor.ReInitialize()

    @Slot()
    def printView(self):
        LOG.debug('LOG.debug view stats')
        fp = self.camera.GetFocalPoint()
        LOG.debug('focal point {}'.format(fp))
        p = self.camera.GetPosition()
        print('position {}'.format(p))
        # dist = math.sqrt( (p[0]-fp[0])**2 + (p[1]-fp[1])**2 + (p[2]-fp[2])**2 )
        # print(dist)
        # self.camera.SetPosition(10, -40, -1)
        # self.setViewUp(0.0, 1.0, 0.0)
        # self.renderer.ResetCamera()
        vu = self.camera.GetViewUp()
        LOG.debug('view up {}'.format(vu))
        d = self.camera.GetDistance()
        print('distance {}'.format(d))
        # self.interactor.ReInitialize()

    @Slot()
    def clearLivePlot(self):
        LOG.debug('clear live plot')
        self.renderer.RemoveActor(self.path_cache_actor)
        self.path_cache_actor = PathCacheActor(self.tooltip_position)
        self.renderer.AddActor(self.path_cache_actor)
        self.renderer_window.Render()

    @Slot(bool)
    def enable_panning(self, enabled):
        self.pan_mode = enabled

    @Slot(bool)
    def enableMultiTouch(self, enabled):
        self.touch_enabled = enabled

    @Slot(bool)
    def setProgramViewWhenLoadingProgram(self, enabled):
        self.touch_enabled = enabled

    @Slot()
    def zoomIn(self):
        if self.camera.GetParallelProjection():
            parallelScale = self.camera.GetParallelScale() * 0.9
            self.camera.SetParallelScale(parallelScale)
            LOG.debug("---camera parallel projection {}".format(parallelScale))
        else:
            self.renderer.ResetCameraClippingRange()
            self.camera.Zoom(1.1)
            LOG.debug("---camera clipping range")

        self.renderer_window.Render()

    @Slot()
    def zoomOut(self):
        if self.camera.GetParallelProjection():
            parallelScale = self.camera.GetParallelScale() * 1.1
            self.camera.SetParallelScale(parallelScale)
        else:
            self.renderer.ResetCameraClippingRange()
            self.camera.Zoom(0.9)

        self.renderer_window.Render()

    @Slot(bool)
    def alphaBlend(self, alpha):
        LOG.debug('alpha blend')

    @Slot(bool)
    @Slot(object)
    def showGrid(self, grid):
        LOG.debug('show grid')
        self.machine_actor.showGridlines(grid)
        self.renderer_window.Render()

    @Slot(bool)
    @Slot(object)
    def showProgramBounds(self, show):
        self.show_program_bounds = show
        for wcs_index, actor in self.path_actors.items():
            program_bounds_actor = self.program_bounds_actors[wcs_index]
            if program_bounds_actor is not None:
                program_bounds_actor.showProgramBounds(self.show_program_bounds)
                self.renderer_window.Render()

    @Slot()
    def toggleProgramBounds(self):
        for wcs_index, actor in self.path_actors.items():
            program_bounds_actor = self.program_bounds_actors[wcs_index]
            self.showProgramBounds(not program_bounds_actor.GetXAxisVisibility())

    @Slot(bool)
    @Slot(object)
    def showProgramTicks(self, ticks):
        for wcs_index, actor in self.path_actors.items():
            program_bounds_actor = self.program_bounds_actors[wcs_index]
            if program_bounds_actor is not None:
                program_bounds_actor.showProgramTicks(ticks)
        self.renderer_window.Render()

    @Slot()
    def toggleProgramTicks(self):
        for wcs_index, actor in self.path_actors.items():
            program_bounds_actor = self.program_bounds_actors[wcs_index]
            self.showProgramTicks(not program_bounds_actor.GetXAxisTickVisibility())

    @Slot(bool)
    @Slot(object)
    def showProgramLabels(self, labels):
        for wcs_index, actor in self.path_actors.items():
            program_bounds_actor = self.program_bounds_actors[wcs_index]
            if program_bounds_actor is not None:
                program_bounds_actor.showProgramLabels(labels)
        self.renderer_window.Render()

    @Slot()
    def toggleProgramLabels(self):
        for wcs_index, actor in self.path_actors.items():
            program_bounds_actor = self.program_bounds_actors[wcs_index]
            self.showProgramLabels(not program_bounds_actor.GetXAxisLabelVisibility())

    @Slot(bool)
    @Slot(object)
    def showMachineBounds(self, bounds):
        self.machine_actor.showMachineBounds(bounds)
        self.renderer_window.Render()

    @Slot()
    def toggleMachineBounds(self):
        self.showMachineBounds(not self.machine_actor.GetXAxisVisibility())

    @Slot(bool)
    @Slot(object)
    def showMachineTicks(self, ticks):
        self.machine_actor.showMachineTicks(ticks)
        self.renderer_window.Render()

    @Slot()
    def toggleMachineTicks(self):
        self.showMachineTicks(not self.machine_actor.GetXAxisTickVisibility())

    @Slot(bool)
    @Slot(object)
    def showMachineLabels(self, labels):
        self.machine_actor.showMachineLabels(labels)
        self.renderer_window.Render()

    @Slot()
    def toggleMachineLabels(self):
        self.showMachineLabels(not self.machine_actor.GetXAxisLabelVisibility())

    @Slot(bool)
    @Slot(object)
    def showMultiColorPath(self, color):
        pass

    @Slot()
    def toggleMultiColorPath(self):
        pass

    @Property(QColor)
    def backgroundColor(self):
        return self._background_color

    @backgroundColor.setter
    def backgroundColor(self, color):
        self._background_color = color

        self.renderer.SetBackground(color.getRgbF()[:3])
        self.renderer_window.Render()

    @backgroundColor.reset
    def backgroundColor(self):
        self._background_color = QColor(0, 0, 0)

        self.renderer.GradientBackgroundOff()
        self.renderer_window.Render()


    @Property(QColor)
    def backgroundColor2(self):
        return self._background_color2

    @backgroundColor2.setter
    def backgroundColor2(self, color2):
        self._background_color2 = color2

        self.renderer.GradientBackgroundOn()
        self.renderer.SetBackground2(color2.getRgbF()[:3])
        self.renderer_window.Render()

    @backgroundColor2.reset
    def backgroundColor2(self):
        self._background_color2 = QColor(0, 0, 0)

        self.renderer.GradientBackgroundOff()
        self.renderer_window.Render()

    @Property(bool)
    def enableProgramTicks(self):
        return self._enableProgramTicks

    @enableProgramTicks.setter
    def enableProgramTicks(self, enable):
        self._enableProgramTicks = enable
    
    # Traverse color property

    @Property(QColor)
    def traverseColor(self):
        return self._traverse_color

    @traverseColor.setter
    def traverseColor(self, color):
        self._traverse_color = color

    @traverseColor.reset
    def traverseColor(self):
        self._traverse_color = self._default_traverse_color
    
    # Arcfeed color property

    @Property(QColor)
    def arcfeedColor(self):
        return self._arcfeed_color

    @arcfeedColor.setter
    def arcfeedColor(self, color):
        self._arcfeed_color = color

    @arcfeedColor.reset
    def arcfeedColor(self):
        self._arcfeed_color = self._default_arcfeed_color
    
    # Feed color property

    @Property(QColor)
    def feedColor(self):
        return self._feed_color

    @feedColor.setter
    def feedColor(self, color):
        self._feed_color = color

    @feedColor.reset
    def feedColor(self):
        self._feed_color = self._default_feed_color
    
    # Dwell color property

    @Property(QColor)
    def dwellColor(self):
        return self._dwel_color

    @dwellColor.setter
    def dwellColor(self, color):
        self._dwel_color = color

    @dwellColor.reset
    def dwellColor(self):
        self._dwel_color = self._default_dwell_color
    
    # User color property

    @Property(QColor)
    def userColor(self):
        return self._user_color

    @userColor.setter
    def userColor(self, color):
        self._user_color = color

    @userColor.reset
    def userColor(self):
        self._user_color = self._default_user_color



"""
GcodeTextEdit
-------------

QPlainTextEdit based G-code editor with syntax highlighting.
"""

import os
import oyaml as yaml

from qtpy.QtCore import (Qt, QRect, QRegularExpression, QEvent, Slot, Signal, Property)
from qtpy.QtGui import (QFont, QColor, QPainter, QSyntaxHighlighter, QTextDocument,
                        QTextOption, QTextFormat, QTextCharFormat, QTextCursor)
from qtpy.QtWidgets import (QApplication, QPlainTextEdit, QTextEdit, QWidget, QMenu, QPlainTextDocumentLayout)

from qtpyvcp import DEFAULT_CONFIG_FILE
from qtpyvcp.plugins import getPlugin
from qtpyvcp.actions import program_actions

STATUS = getPlugin('status')
YAML_DIR = os.path.dirname(DEFAULT_CONFIG_FILE)


class GcodeSyntaxHighlighter(QSyntaxHighlighter):
    def __init__(self, parent):
        super(GcodeSyntaxHighlighter, self).__init__(parent.document())

        self._parent = parent

        self.rules = []
        self.char_fmt = QTextCharFormat()

        self._abort = False

        self.loadSyntaxFromYAML()

    def loadSyntaxFromYAML(self):

        with open(os.path.join(YAML_DIR, 'gcode_syntax.yml')) as fh:
            syntax_specs = yaml.load(fh, Loader=yaml.FullLoader)

        assert isinstance(syntax_specs, dict), \
            "Invalid YAML format for language spec, root item must be a dictionary."

        cio = QRegularExpression.CaseInsensitiveOption

        for lang_name, language in syntax_specs.items():

            definitions = language.get('definitions', {})

            default_fmt_spec = definitions.get('default', {}).get('textFormat', {})

            for context_name, spec in definitions.items():

                base_fmt = default_fmt_spec.copy()
                fmt_spec = spec.get('textFormat', {})

                # update the default fmt spec
                base_fmt.update(fmt_spec)

                char_fmt = self.charFormatFromSpec(fmt_spec)

                patterns = spec.get('match', [])
                for pattern in patterns:
                    self.rules.append([QRegularExpression(pattern, cio), char_fmt])

    def charFormatFromSpec(self, fmt_spec):

        char_fmt = self.defaultCharFormat()

        for option, value in fmt_spec.items():
            if value is None:
                continue

            if option in ['foreground', 'background']:
                value = QColor(value)

            if isinstance(value, basestring) and value.startswith('QFont:'):
                value = getattr(QFont, value[6:])

            attr = 'set' + option[0].capitalize() + option[1:]
            getattr(char_fmt, attr)(value)

        return char_fmt

    def defaultCharFormat(self):
        char_fmt = QTextCharFormat()
        char_fmt.setFont(self._parent.font())
        return char_fmt

    def highlightBlock(self, text):
        """Apply syntax highlighting to the given block of text.
        """

        QApplication.processEvents()

        for regex, fmt in self.rules:

            nth = 0
            match = regex.match(text, offset=0)
            index = match.capturedStart()

            while index >= 0:

                # We actually want the index of the nth match
                index = match.capturedStart(nth)
                length = match.capturedLength(nth)
                self.setFormat(index, length, fmt)

                # check the rest of the string
                match = regex.match(text, offset=index + length)
                index = match.capturedStart()


class GcodeTextEdit(QPlainTextEdit):
    """G-code Text Edit

    QPlainTextEdit based G-code editor with syntax heightening.
    """
    focusLine = Signal(int)

    def __init__(self, parent=None):
        super(GcodeTextEdit, self).__init__(parent)

        self.setCenterOnScroll(True)
        self.setGeometry(50, 50, 800, 640)
        self.setWordWrapMode(QTextOption.NoWrap)

        self.block_number = None
        self.focused_line = 1
        self.current_line_background = QColor(self.palette().alternateBase())

        self.old_docs = []
        # set the custom margin
        self.margin = NumberMargin(self)

        # set the syntax highlighter
        self.gCodeHighlighter = GcodeSyntaxHighlighter(self)

        # context menu
        self.menu = QMenu(self)
        self.menu.addAction(self.tr("Run from line {}".format(self.focused_line)), self.runFromHere)
        self.menu.addSeparator()
        self.menu.addAction(self.tr('Cut'), self.cut)
        self.menu.addAction(self.tr('Copy'), self.copy)
        self.menu.addAction(self.tr('Paste'), self.paste)

        # FixMe: Picks the first action run from here, should not be by index
        self.run_action = self.menu.actions()[0]
        self.run_action.setEnabled(program_actions.run_from_line.ok())
        program_actions.run_from_line.bindOk(self.run_action)

        # connect signals
        self.cursorPositionChanged.connect(self.onCursorChanged)

        # connect status signals
        STATUS.file.notify(self.loadProgramFile)
        STATUS.motion_line.onValueChanged(self.setCurrentLine)

    def keyPressEvent(self, event):
        # keep the cursor centered
        if event.key() == Qt.Key_Up:
            self.moveCursor(QTextCursor.Up)
            self.centerCursor()

        elif event.key() == Qt.Key_Down:
            self.moveCursor(QTextCursor.Down)
            self.centerCursor()

        else:
            super(GcodeTextEdit, self).keyPressEvent(event)

    def changeEvent(self, event):
        if event.type() == QEvent.FontChange:
            # Update syntax highlighter with new font
            self.gCodeHighlighter = GcodeSyntaxHighlighter(self)
        super(GcodeTextEdit, self).changeEvent(event)

    def setPlainText(self, p_str):
        # FixMe: Keep a reference to old QTextDocuments form previously loaded
        # files. This is needed to prevent garbage collection which results in a
        # seg fault if the document is discarded while still being highlighted.
        self.old_docs.append(self.document())

        doc = QTextDocument()
        doc.setDocumentLayout(QPlainTextDocumentLayout(doc))
        doc.setPlainText(p_str)

        self.setDocument(doc)
        self.margin.updateWidth()

        # start syntax heightening
        self.gCodeHighlighter = GcodeSyntaxHighlighter(self)

    @Slot(bool)
    def EditorReadOnly(self, state):
        """Set to Read Only to disable editing"""
        if state:
            self.setReadOnly(True)
        else:
            self.setReadOnly(False)

    @Property(QColor)
    def currentLineBackground(self):
        return self.current_line_background

    @currentLineBackground.setter
    def currentLineBackground(self, color):
        self.current_line_background = color
        # Hack to get background to update
        self.setCurrentLine(2)
        self.setCurrentLine(1)

    @Property(QColor)
    def marginBackground(self):
        return self.margin.background

    @marginBackground.setter
    def marginBackground(self, color):
        self.margin.background = color
        self.margin.update()

    @Property(QColor)
    def marginCurrentLineBackground(self):
        return self.margin.highlight_background

    @marginCurrentLineBackground.setter
    def marginCurrentLineBackground(self, color):
        self.margin.highlight_background = color
        self.margin.update()

    @Property(QColor)
    def marginColor(self):
        return self.margin.color

    @marginColor.setter
    def marginColor(self, color):
        self.margin.color = color
        self.margin.update()

    @Property(QColor)
    def marginCurrentLineColor(self):
        return self.margin.highlight_color

    @marginCurrentLineColor.setter
    def marginCurrentLineColor(self, color):
        self.margin.highlight_color = color
        self.margin.update()

    @Slot(str)
    @Slot(object)
    def loadProgramFile(self, fname=None):
        if fname:
            with open(fname) as f:
                gcode = f.read()
            self.setPlainText(gcode)

    @Slot(int)
    @Slot(object)
    def setCurrentLine(self, line):
        cursor = QTextCursor(self.document().findBlockByLineNumber(line - 1))
        self.setTextCursor(cursor)
        self.centerCursor()

    def getCurrentLine(self):
        return self.textCursor().blockNumber() + 1

    def onCursorChanged(self):
        # highlights current line, find a way not to use QTextEdit
        block_number = self.textCursor().blockNumber()
        if block_number != self.block_number:
            self.block_number = block_number
            selection = QTextEdit.ExtraSelection()
            selection.format.setBackground(self.current_line_background)
            selection.format.setProperty(QTextFormat.FullWidthSelection, True)
            selection.cursor = self.textCursor()
            selection.cursor.clearSelection()
            self.setExtraSelections([selection])

        # emit signals for backplot etc.
        self.focused_line = block_number + 1
        self.focusLine.emit(self.focused_line)

    def contextMenuEvent(self, event):
        self.run_action.setText("Run from line {}".format(self.focused_line))
        self.menu.popup(event.globalPos())
        event.accept()

    def runFromHere(self, *args, **kwargs):
        line = self.getCurrentLine()
        program_actions.run(line)

    def resizeEvent(self, *e):
        cr = self.contentsRect()
        rec = QRect(cr.left(), cr.top(), self.margin.getWidth(), cr.height())
        self.margin.setGeometry(rec)
        QPlainTextEdit.resizeEvent(self, *e)


class NumberMargin(QWidget):
    def __init__(self, parent):
        QWidget.__init__(self, parent)
        self.parent = parent
        # this only happens when lines are added or subtracted
        self.parent.blockCountChanged.connect(self.updateWidth)
        # this happens quite often
        self.parent.updateRequest.connect(self.updateContents)

        self.background = QColor('#e8e8e8')
        self.highlight_background = QColor('#e8e8e8')
        self.color = QColor('#717171')
        self.highlight_color = QColor('#000000')

    def getWidth(self):
        blocks = self.parent.blockCount()
        return self.parent.fontMetrics().width(str(blocks)) + 5

    def updateWidth(self): # check the number column width and adjust
        width = self.getWidth()
        if self.width() != width:
            self.setFixedWidth(width)
            self.parent.setViewportMargins(width, 0, 0, 0)

    def updateContents(self, rect, scroll):
        if scroll:
            self.scroll(0, scroll)
        else:
            self.update(0, rect.y(), self.width(), rect.height())

        if rect.contains(self.parent.viewport().rect()):
            self.updateWidth()

    def paintEvent(self, event):  # this puts the line numbers in the margin
        painter = QPainter(self)
        painter.fillRect(event.rect(), self.background)
        block = self.parent.firstVisibleBlock()

        font = self.parent.font()

        while block.isValid():
            block_num = block.blockNumber()
            block_top = self.parent.blockBoundingGeometry(block).translated(self.parent.contentOffset()).top()

            # if the block is not visible stop wasting time
            if not block.isVisible() or block_top >= event.rect().bottom():
                break

            if block_num == self.parent.textCursor().blockNumber():
                font.setBold(True)
                painter.setFont(font)
                painter.setPen(self.highlight_color)
                background = self.highlight_background
            else:
                font.setBold(False)
                painter.setFont(font)
                painter.setPen(self.color)
                background = self.background

            text_rec = QRect(0, block_top, self.width(), self.parent.fontMetrics().height())
            painter.fillRect(text_rec, background)
            painter.drawText(text_rec, Qt.AlignRight, str(block_num + 1))
            block = block.next()

        painter.end()
        QWidget.paintEvent(self, event)



import linuxcnc
import os

from qtpy.QtCore import Signal
from qtpyvcp.plugins import getPlugin
from qtpyvcp.utilities import logger
from qtpyvcp.utilities.info import Info

INFO = Info()
LOG = logger.getLogger(__name__)
from PyQt5.QtCore import QObject

IN_DESIGNER = os.getenv('DESIGNER', False)
"""
This class acts as a datasource for the VTK components.
It abstracts all the linuxcnc specific logic and exposes simple methods that could be eventually 
mocked for testing the VTK outside of a linuxcnc context.
"""
class LinuxCncDataSource(QObject):
    programLoaded = Signal(str)
    positionChanged = Signal(tuple)
    motionTypeChanged = Signal(int)
    g5xOffsetChanged = Signal(tuple)
    g92OffsetChanged = Signal(tuple)
    g5xIndexChanged = Signal(int)
    offsetTableChanged = Signal(dict)
    activeOffsetChanged = Signal(int)
    toolTableChanged = Signal(tuple)
    toolOffsetChanged = Signal(tuple)

    def __init__(self):
        super(LinuxCncDataSource, self).__init__(None)

        self._info = Info()
        self._status = getPlugin('status')
        self._tooltable = getPlugin('tooltable')
        self._offsettable = getPlugin('offsettable')
        self._inifile = linuxcnc.ini(os.getenv("INI_FILE_NAME"))
        self._is_lathe = bool(self._inifile.find("DISPLAY", "LATHE"))

        self._status.file.notify(self.__handleProgramLoaded)
        self._status.position.notify(self.__handlePositionChanged)
        self._status.motion_type.notify(self.__handleMotionTypeChanged)
        self._status.g5x_offset.notify(self.__handleG5xOffsetChange)
        self._status.g92_offset.notify(self.__handleG92OffsetChange)

        self._status.g5x_index.notify(self.__handleG5xIndexChange)
        self._status.rotation_xy.notify(self.__handleRotationChangeXY)

        self._offsettable.offset_table_changed.connect(self.__handleOffsetTableChanged)
        self._offsettable.active_offset_changed.connect(self.__handleActiveOffsetChanged)

        self._status.tool_offset.notify(self.__handleToolOffsetChanged)
        self._status.tool_table.notify(self.__handleToolTableChanged)

    def __handleProgramLoaded(self, fname):
        #LOG.debug("__handleProgramLoaded: {}".format(fname))
        self.programLoaded.emit(fname)

    def __handlePositionChanged(self, position):
        #LOG.debug("__handlePositionChanged: {}".format(type(position)))
        self.positionChanged.emit(position)

    def __handleMotionTypeChanged(self, motion_type):
        #LOG.debug("__handleMotionTypeChanged: {}".format(motion_type))
        self.motionTypeChanged.emit(motion_type)

    def __handleG5xOffsetChange(self, offset):
        # the received parameter, its missing the rotation of the current wcs
        emitted_offset = list(offset)
        active_wcs = self.getWcsOffsets()[self.getActiveWcsIndex()]

        LOG.debug("--------initial offset emitted: {} {}".format(type(offset),offset))
        LOG.debug("--------active wcs: {} {}".format(type(active_wcs), active_wcs))

        emitted_offset.append(self.__getRotationOfActiveWcs())
        LOG.debug("--------correct_offset: {}".format(emitted_offset))
        result = tuple(emitted_offset)
        LOG.debug("--------result: {} {}".format(type(result), result))
        self.g5xOffsetChanged.emit(result)

    def __handleG92OffsetChange(self, offset):
        #LOG.debug("__handleG92OffsetChange: {}".format(type(offset)))
        self.g92OffsetChanged.emit(offset)

    def __handleG5xIndexChange(self, value):
        LOG.debug("__handleG5xIndexChange: {}".format(value - 1))
        self.g5xIndexChanged.emit(value - 1)

    def __handleRotationChangeXY(self, value):
        LOG.debug("__handleRotationChangeXY: {}".format(value))
        active_wcs = self.getWcsOffsets()[self.getActiveWcsIndex()]
        LOG.debug("--------active wcs: {} {}".format(type(active_wcs), active_wcs))
        active_wcs[9] = value
        LOG.debug("--------active new wcs: {} {}".format(type(active_wcs), active_wcs))
        self.g5xOffsetChanged.emit(tuple(active_wcs))

    def __handleOffsetTableChanged(self, offset_table):
        #LOG.debug("__handleOffsetTableChanged: {}".format(type(offset_table)))
        self.offsetTableChanged.emit(offset_table)
        #self.g5xOffsetChanged.emit(self.getActiveWcsOffsets())

    def __handleActiveOffsetChanged(self, active_offset_index):
        # the first one is g53 - machine coordinates, we're not interested in that one
        current_wcs_index = active_offset_index - 1
        LOG.debug("__handleActiveOffsetChanged index: {}".format(current_wcs_index))
        self.activeOffsetChanged.emit(current_wcs_index)

    def __handleToolOffsetChanged(self, tool_offset):
        #LOG.debug("__handleToolOffsetChanged: {}".format(type(tool_offset)))
        self.toolOffsetChanged.emit(tool_offset)

    def __handleToolTableChanged(self, tool_table):
        #LOG.debug("__handleToolTableChanged: {}".format(type(tool_table)))
        self.toolTableChanged.emit(tool_table)

    def getAxis(self):
        return self._status.stat.axis

    def getAxisMask(self):
        return self._status.stat.axis_mask

    def getToolTable(self):
        return self._status.stat.tool_table

    def getToolOffset(self):
        return self._status.tool_offset

    def isMachineMetric(self):
        return self._info.getIsMachineMetric()

    def getProgramUnits(self):
        return str(self._status.program_units)

    def isMachineLathe(self):
        return self._is_lathe

    def isModeMdi(self):
        return str(self._status.task_mode) == "MDI"

    def isModeAuto(self):
        return str(self._status.task_mode) == "Auto"

    def getActiveWcsIndex(self):
        # in the stat, the first one the list is G53 (Machine Coordinates)
        # therefore to get the correct index of the G54 we need to do a -1
        return self._status.stat.g5x_index - 1

    def getActiveWcsOffsets(self):
        # g5x_offset does not contain the rotation information
        xx = self._status.stat.g5x_offset
        LOG.debug("self._status.stat.g5x_offset: {}".format(type(xx)))
        xy = list(xx)
        xy.append(self.__getRotationOfActiveWcs())
        return tuple(xy)

    def __getRotationOfActiveWcs(self):
        if not IN_DESIGNER:
            current_wcs = self.getWcsOffsets()[self.getActiveWcsIndex()]
            LOG.debug("-----current_wcs index: {}".format(current_wcs))
            return current_wcs[9]
        else:
            return 0

    def getG92_offset(self):
        return self._status.stat.g92_offset

    def getWcsOffsets(self):
        # returns a dictionary with the coordinate systems from 0 to 8 (g54 up to g59.3)
        return self._offsettable.getOffsetTable()



from qtpy.QtCore import Qt, Slot, Property, QStringListModel
from qtpy.QtGui import QValidator
from qtpy.QtWidgets import QLineEdit, QListWidgetItem, QCompleter

from qtpyvcp.plugins import getPlugin
from qtpyvcp.utilities.info import Info
from qtpyvcp.actions.machine_actions import issue_mdi
from qtpyvcp.widgets.base_widgets.base_widget import CMDWidget

STATUS = getPlugin('status')
INFO = Info()


class Validator(QValidator):
    def validate(self, string, pos):
        # eventually could do some actual validating here
        inputVal = string
        if not(inputVal.startswith(";") or inputVal.startswith("(")):
            inputVal = inputVal.upper();
        return QValidator.Acceptable, inputVal, pos


class MDIEntry(QLineEdit, CMDWidget):
    """MDI Entry

    Input any valid g Code. Enter sends the g Code.
    """
    def __init__(self, parent=None):
        super(MDIEntry, self).__init__(parent)

        self.mdi_rtnkey_behaviour_supressed = False
        self.mdi_history_size = 100

        self.validator = Validator(self)
        self.setValidator(self.validator)
        
        # The completer is what creates the visual list mdi history
        # that can be selected from. This is useful when mdientry is used
        # stand alone but can get in the way when using this widget in
        # conjunction with mdihistory widget and a rich on-screen data
        # entry UI. So this designer exposed setting allows this to be enabled
        # or disabled as needed.
        self._completer_enabled = True

        self.returnPressed.connect(self.submit)

    @Property(int)
    def mdi_history_size(self):
        return self._mdi_history_size

    @mdi_history_size.setter
    def mdi_history_size(self, size):
        self._mdi_history_size = size

    @Property(bool)
    def completerEnabled(self):
        return self._completer_enabled
    
    @completerEnabled.setter
    def completerEnabled(self, flag):
        self._completer_enabled = flag

    @Slot()
    def submit(self):
        # Only support submit from Return Key if not suppressed
        # This is used to stop standalone behaviour by the MDIHistory widget.
        # MDIHisttory uses its handle to this widget to set
        # mdi_rtnkey_behaviour_supressed = True.
        # MDIHistory will now manage the MDILine submission process
        if not self.mdi_rtnkey_behaviour_supressed:
            cmd = str(self.text()).strip()
            issue_mdi(cmd)
            self.setText('')
            STATUS.mdi_history.setValue(cmd)

    @Slot(QListWidgetItem)
    def setMDIText(self, listItem):
        if listItem is not None:
            self.setText(listItem.text())

    def keyPressEvent(self, event):
        if (    self._completer_enabled
             and
             (  event.key() == Qt.Key_Up
              or event.key() == Qt.Key_Down
             )
           ):
            self.completer().complete()
        else:
            super(MDIEntry, self).keyPressEvent(event)

    def focusInEvent(self, event):
        super(MDIEntry, self).focusInEvent(event)
        if self._completer_enabled:
            self.completer().complete()

    def supress_rtn_key_behaviour(self):
        self.mdi_rtnkey_behaviour_supressed = True

    def enable_rtn_key_behaviour(self):
        self.mdi_rtnkey_behaviour_supressed = False

    def initialize(self):
        history = STATUS.mdi_history.value
        if self._completer_enabled:
            completer = QCompleter()
            completer.setCaseSensitivity(Qt.CaseInsensitive)
            self.model = QStringListModel()
            completer.setModel(self.model)
            self.setCompleter(completer)
            self.model.setStringList(history)
            STATUS.mdi_history.notify(self.model.setStringList)

        STATUS.max_mdi_history_length = self.mdi_history_size

    def terminate(self):
        pass



import os

import linuxcnc
from qtpy import uic, QtWidgets
from qtpy.QtCore import Slot, Qt, QEvent, QPoint
from qtpy.QtGui import QInputMethodEvent, QGuiApplication, QKeyEvent
from qtpy.QtWidgets import QWidget, QAbstractButton, QAbstractSpinBox


class VirtualInput(QWidget):
    """VirtualInput

    This class provides support for a virtual num pad and keyboard.
    If the TOUCHSCREEN value is set to True, 1, or yes in the DISPLAY
    section of the INI file, the input panel will be shown for controls
    that can be edited.

    At the moment the panel to show is chosen heuristically if the control
    has text, and setText methods. The num pad is shown if the control is a SpinBox.
    """

    def __init__(self, ui_file, parent=None):
        super(VirtualInput, self).__init__(parent)

        self.caps_on = False
        self.focus_object = None
        self.setWindowFlags(Qt.Tool | Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint | Qt.BypassWindowManagerHint)
        uic.loadUi(ui_file, self)

    @Slot(QAbstractButton)
    def on_buttonGroup_buttonPressed(self, btn):
        event = QInputMethodEvent()
        event.setCommitString(btn.text())
        QGuiApplication.sendEvent(self.focus_object, event)

    @Slot()
    def on_space_key_pressed(self):
        event = QInputMethodEvent()
        event.setCommitString(' ')
        QGuiApplication.sendEvent(self.focus_object, event)

    @Slot()
    def on_enter_key_pressed(self):
        self.send_key_(Qt.Key_Enter)

    @Slot()
    def on_esc_key_pressed(self):
        self.send_key_(Qt.Key_Escape)
        self.hide()

    @Slot()
    def on_backspace_key_pressed(self):
        self.send_key_(Qt.Key_Backspace)

    @Slot()
    def on_left_key_pressed(self):
        self.send_key_(Qt.Key_Left)

    @Slot()
    def on_right_key_pressed(self):
        self.send_key_(Qt.Key_Right)

    @Slot()
    def on_up_key_pressed(self):
        self.send_key_(Qt.Key_Up)

    @Slot()
    def on_down_key_pressed(self):
        self.send_key_(Qt.Key_Down)

    @Slot()
    def on_tab_key_pressed(self):
        self.send_key_(Qt.Key_Tab)

    @Slot()
    def on_home_key_pressed(self):
        self.send_key_(Qt.Key_Home)

    @Slot()
    def on_end_key_pressed(self):
        self.send_key_(Qt.Key_End)

    @Slot()
    def on_page_up_key_pressed(self):
        self.send_key_(Qt.Key_PageUp)

    @Slot()
    def on_page_down_key_pressed(self):
        self.send_key_(Qt.Key_PageDown)

    @Slot()
    def on_caps_key_pressed(self):
        self.caps_on = not self.caps_on
        for btn in self.buttonGroup.buttons():
            if self.caps_on:
                btn.setText(btn.text().upper())
            else:
                btn.setText(btn.text().lower())

    def set_focus_object(self, value):
        self.focus_object = value
        self.position_keyboard_to_object(self.focus_object)

    def activate(self, obj):
        self.set_focus_object(obj)
        self.show()

    def position_keyboard_to_object(self, obj):
        screen = QtWidgets.QDesktopWidget().screenGeometry()
        pos = obj.mapToGlobal(QPoint(0, obj.height()))

        if pos.y() + self.height() > screen.height():
            pos.setY(pos.y() - self.height() - obj.height())
        if pos.x() + self.width() > screen.width():
            pos.setX(pos.x() - self.width() + obj.width())
        if pos.x() + self.width() > screen.width():
            pos.setX(pos.x() - (pos.x() + self.width() - screen.width()))

        self.move(pos.x(), pos.y())

    def send_key_(self, key):
        event = QKeyEvent(QEvent.KeyPress, key, Qt.NoModifier)
        QGuiApplication.postEvent(self.focus_object, event)

        event = QKeyEvent(QEvent.KeyRelease, key, Qt.NoModifier)
        QGuiApplication.postEvent(self.focus_object, event)
